#!/usr/bin/env python3

import dataclasses
import enum
import json
import logging
import os
import sys
from argparse import FileType, Namespace
from dataclasses import asdict, dataclass, fields, make_dataclass
from typing import Sequence

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(name)s | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=os.environ.get("LOGLEVEL", "INFO").upper(),
    stream=sys.stdout,
)
logger = logging.getLogger(__name__)

import simple_parsing
import torch
from simple_parsing import choice, field, flag
from simple_parsing.wrappers import dataclass_wrapper
from tabulate import tabulate, tabulate_formats
from tqdm import tqdm

from mbrs import registry, timer
from mbrs.args import ArgumentParser, DataclassWrapper
from mbrs.decoders import (
    DecoderBase,
    DecoderReferenceBased,
    DecoderReferenceless,
    get_decoder,
)
from mbrs.metrics import Metric, MetricEnum, get_metric
from mbrs.selectors import Selector, get_selector

simple_parsing.parsing.logger.setLevel(logging.ERROR)
dataclass_wrapper.logger.setLevel(logging.ERROR)


class Format(enum.Enum):
    plain = "plain"
    json = "json"


@dataclass
class CommonArguments:
    """Common arguments."""

    # Hypotheses file.
    hypotheses: str = field(positional=True)
    # Number of candidates.
    num_candidates: int = field(alias=["-n"])
    # Source file.
    source: str | None = field(default=None, alias=["-s"])
    # References file.
    references: str | None = field(default=None, alias=["-r"])
    # References log-probabilities file.
    reference_lprobs: str | None = field(default=None)
    # Output file.
    output: FileType("w", encoding="utf-8") = field(default="-", alias=["-o"])
    # Output format.
    format: Format = choice(Format, default=Format.plain)
    # Number of references for each sentence.
    num_references: int | None = field(default=None)
    # Type of the decoder.
    decoder: str = choice(*registry.get_registry("decoder").keys(), default="mbr")
    # Type of the metric.
    metric: str = choice(*registry.get_registry("metric").keys(), default="bleu")
    # Type of the selector.
    selector: str = choice(*registry.get_registry("selector").keys(), default="nbest")
    # Return the n-best hypotheses.
    nbest: int = field(default=1)
    # No verbose information and report.
    quiet: bool = flag(default=False)
    # Report file.
    report: FileType("w") = field(default="-")
    # Report runtime statistics with the given format.
    report_format: str = choice(*tabulate_formats, default="rounded_outline")
    # Number of digits for values of float point.
    width: int = field(default=1, alias=["-w"])


def get_argparser(args: Sequence[str] | None = None) -> ArgumentParser:
    meta_parser = ArgumentParser(add_help=False, add_config_path_arg=True)
    meta_parser.add_arguments(
        CommonArguments, "common", dataclass_wrapper_class=DataclassWrapper
    )
    for _field in meta_parser._wrappers[0].fields:
        _field.required = False
    known_args, _ = meta_parser.parse_known_args(args=args)
    metric_type = get_metric(known_args.common.metric)
    decoder_type = get_decoder(known_args.common.decoder)
    selector_type = get_selector(known_args.common.selector)

    parser = ArgumentParser(add_help=False, add_config_path_arg=True)
    parser.add_arguments(
        CommonArguments, "common", dataclass_wrapper_class=DataclassWrapper
    )
    parser.add_arguments(metric_type.Config, "metric", prefix="metric.")
    parser.add_arguments(decoder_type.Config, "decoder", prefix="decoder.")
    parser.add_arguments(selector_type.Config, "selector", prefix="selector.")
    for _field in parser._wrappers[0].fields:
        _field.required = False

    known_args, _ = parser.parse_known_args(args=args)
    for cfg, m in [
        (known_args.metric, metric_type),
        (known_args.decoder, decoder_type),
        (known_args.selector, selector_type),
    ]:
        for _field in fields(cfg):
            field_name = _field.name
            field_attr = getattr(cfg, field_name)
            if isinstance(field_attr, MetricEnum):
                config_type = get_metric(field_attr).Config
                m.Config = make_dataclass(
                    m.Config.__name__,
                    fields=[
                        (
                            field_name,
                            type(field_attr),
                            dataclasses.field(default=field_attr),
                        ),
                        (
                            field_name + "_config",
                            config_type,
                            dataclasses.field(default_factory=config_type),
                        ),
                    ],
                    bases=(m.Config,),
                )

    parser = ArgumentParser(add_help=True, add_config_path_arg=True)
    parser.add_arguments(
        CommonArguments, "common", dataclass_wrapper_class=DataclassWrapper
    )
    parser.add_arguments(metric_type.Config, "metric", prefix="metric.")
    parser.add_arguments(decoder_type.Config, "decoder", prefix="decoder.")
    parser.add_arguments(selector_type.Config, "selector", prefix="selector.")
    return parser


def format_argparser() -> ArgumentParser:
    parser = get_argparser()
    parser.preprocess_parser()
    return parser


def main(args: Namespace) -> None:
    if not args.common.quiet:
        logger.info(args)

    sources = None
    if args.common.source is not None:
        with open(args.common.source, mode="r") as f:
            sources = f.readlines()

    with open(args.common.hypotheses, mode="r") as f:
        hypotheses = f.readlines()

    references = None
    if args.common.references is not None:
        with open(args.common.references, mode="r") as f:
            references = f.readlines()
    else:
        references = hypotheses

    reference_lprobs = None
    if args.common.reference_lprobs is not None:
        with open(args.common.reference_lprobs, mode="r") as f:
            reference_lprobs = f.readlines()
        assert len(references) == len(reference_lprobs)

    metric_type = get_metric(args.common.metric)
    metric: Metric = metric_type(args.metric)

    selector_type = get_selector(args.common.selector)
    selector: Selector = selector_type(args.selector)

    decoder_type = get_decoder(args.common.decoder)
    decoder: DecoderReferenceBased | DecoderReferenceless = decoder_type(
        args.decoder, metric, selector
    )

    num_cands = args.common.num_candidates
    num_refs = args.common.num_references or num_cands
    num_sents = len(hypotheses) // num_cands
    assert num_sents * num_cands == len(hypotheses)

    def output_results(res: DecoderBase.Output):
        if args.common.format == Format.plain:
            for sent in res.sentence:
                print(sent, file=args.common.output)
        elif args.common.format == Format.json:
            for i, (sent, idx, score) in enumerate(
                zip(res.sentence, res.idx, res.score)
            ):
                print(
                    json.dumps(
                        {
                            "rank": i,
                            "sentence": sent,
                            "selected_idx": idx,
                            "expected_score": score,
                            **{
                                k: v
                                for k, v in asdict(res).items()
                                if k not in {"sentence", "idx", "score"}
                            },
                        },
                        ensure_ascii=False,
                    ),
                    file=args.common.output,
                )

    if isinstance(decoder, DecoderReferenceless):
        for i in tqdm(range(num_sents)):
            src = sources[i].strip()
            hyps = [h.strip() for h in hypotheses[i * num_cands : (i + 1) * num_cands]]
            with timer.measure("total"):
                output = decoder.decode(hyps, src, args.common.nbest)
            output_results(output)
    else:
        for i in tqdm(range(num_sents)):
            src = sources[i].strip() if sources is not None else None
            hyps = [h.strip() for h in hypotheses[i * num_cands : (i + 1) * num_cands]]
            refs = [r.strip() for r in references[i * num_refs : (i + 1) * num_refs]]
            ref_lprobs = None
            if reference_lprobs is not None:
                ref_lprobs = [
                    float(r.strip())
                    for r in reference_lprobs[i * num_refs : (i + 1) * num_refs]
                ]

                # Deduplicate the same elements for the model-based estimation.
                # Note that we regard pairs of `(reference, log-prob of reference)` that are
                # equal as the same samples.
                # We use `dict.fromkeys()` instead of `set()`; thus, the order is kept.
                uniq_refs, uniq_ref_lprobs = tuple(
                    zip(*dict.fromkeys(zip(refs, ref_lprobs)).keys())
                )

                refs = list(uniq_refs)
                ref_lprobs = torch.tensor(
                    list(uniq_ref_lprobs), dtype=torch.float32, device=metric.device
                )

            with timer.measure("total"):
                output = decoder.decode(
                    hyps,
                    refs,
                    src,
                    nbest=args.common.nbest,
                    reference_lprobs=ref_lprobs,
                )
            output_results(output)

    if not args.common.quiet:
        statistics = timer.aggregate().result(num_sents)
        table = tabulate(
            statistics,
            headers="keys",
            tablefmt=args.common.report_format,
            floatfmt=f".{args.common.width}f",
        )
        print(table, file=args.common.report)


def cli_main():
    args = get_argparser().parse_args()
    main(args)


if __name__ == "__main__":
    cli_main()



from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import Optional

import torch
from torch import Tensor


class Selector(abc.ABC):
    """Selector base class."""

    def __init__(self, cfg: Selector.Config) -> None:
        self.cfg = cfg

    @dataclass
    class Config:
        """Configuration for the selector."""

    @dataclass
    class Output:
        """
        - idx (list[int]): Index numbers of the n-best hypotheses.
        - sentence (list[str]): Sentences of the n-best hypotheses.
        - score (list[float]): Scores of the n-best hypotheses.
        """

        idx: list[int]
        sentence: list[str]
        score: list[float]

    def topk(
        self, x: Tensor, k: int = 1, maximize: bool = True
    ) -> tuple[list[float], list[int]]:
        """Return the top-k best elements and corresponding indices.

        Args:
            x (Tensor): Input 1-D array.
            k (int): Return the top-k values and indices.
            maximize (bool): Whether maximize the scores or not.

        Returns:
            tuple[list[float], list[int]]
              - list[float]: The top-k values.
              - list[int]: The top-k indices.
        """
        values, indices = torch.topk(x, k=min(k, len(x)), largest=maximize)
        return values.tolist(), indices.tolist()

    def argbest(self, x: Tensor, maximize: bool = True) -> Tensor:
        """Return the index of the best element.

        Args:
            x (Tensor): Input 1-D array.
            maximize (bool): Whether maximize the scores or not.

        Returns:
            Tensor: A scalar tensor of the best index.
        """
        if maximize:
            return torch.argmax(x)
        return torch.argmin(x)

    def superior(self, a: float, b: float, maximize: bool = True) -> bool:
        """Return whether the score `a` is superior to the score `b`.

        Args:
            a (float): A score.
            b (float): A score.
            maximize (bool): Whether maximize the scores or not.

        Returns:
            bool: Return True when `a` is superior to `b`.
        """
        if maximize:
            return a > b
        return a < b

    @abc.abstractmethod
    def select(
        self,
        hypotheses: list[str],
        expected_scores: Tensor,
        nbest: int = 1,
        source: Optional[str] = None,
        maximize: bool = True,
        *args,
        **kwargs,
    ) -> Selector.Output:
        """Select the final output list.

        Args:
            hypotheses (list[str]): Hypotheses.
            expected_scores (Tensor): The expected scores for each hypothesis.
            nbest (int): Return the n-best hypotheses based on the selection rule.
            source (str, optional): A source.
            maximize (bool): Whether maximize the scores or not.

        Returns:
            Selector.Output: Selected hypotheses.
        """



import pytest
import torch

from mbrs.metrics import get_metric
from mbrs.selectors import Selector

from .aggregate_mbr import DecoderAggregateMBR

HYPOTHESES = [
    ["another test", "this is a test", "this is an test", "this is a fest"],
    ["another test", "this is a fest", "this is a test"],
    ["this is a test"],
    ["Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;"],
]
REFERENCES = [
    ["another test", "this is a test", "this is an test", "this is a fest"],
    ["this is a test", "ref1 ref2 ref3 ref4"],
    ["this is a test"],
    ["producţia de zahăr brut se exprimă în zahăr alb;"],
]

BEST_INDICES_CHRF = [2, 2, 0, 0]
BEST_SENTENCES_CHRF = [
    "this is an test",
    "this is a test",
    "this is a test",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
SCORES_CHRF = torch.Tensor([63.75969, 40.86618, 100.0, 46.16121])

BEST_INDICES_BLEU = [1, 2, 0, 0]
BEST_SENTENCES_BLEU = [
    "this is a test",
    "this is a test",
    "this is a test",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
SCORES_BLEU = {
    True: torch.Tensor([42.36951, 50.0, 100.0, 8.49310]),
    False: torch.Tensor([42.36951, 50.0, 100.0, 8.49310]),
}


class TestDecoderAggregateMBR:
    @pytest.mark.parametrize("nbest", [1, 2])
    def test_decode_chrf(self, nbest: int):
        metric_type = "chrf"
        metric_cls = get_metric(metric_type)
        decoder = DecoderAggregateMBR(
            DecoderAggregateMBR.Config(), metric_cls(metric_cls.Config())
        )
        for i, (hyps, refs) in enumerate(zip(HYPOTHESES, REFERENCES)):
            output = decoder.decode(hyps, refs, nbest=nbest, reference_lprobs=None)
            assert output.idx[0] == BEST_INDICES_CHRF[i]
            assert output.sentence[0] == BEST_SENTENCES_CHRF[i]
            assert len(output.sentence) == min(nbest, len(hyps))
            assert len(output.score) == min(nbest, len(hyps))
            torch.testing.assert_close(
                torch.tensor(output.score[0]),
                SCORES_CHRF[i],
                atol=0.0005,
                rtol=1e-4,
            )

            output = decoder.decode(
                hyps,
                refs,
                nbest=1,
                reference_lprobs=torch.Tensor([-2.000]).repeat(len(refs)),
            )
            assert output.idx[0] == BEST_INDICES_CHRF[i]
            assert output.sentence[0] == BEST_SENTENCES_CHRF[i]
            torch.testing.assert_close(
                torch.tensor(output.score[0]),
                SCORES_CHRF[i],
                atol=0.0005,
                rtol=1e-4,
            )

    @pytest.mark.parametrize("effective_order", [True, False])
    @pytest.mark.parametrize("nbest", [1, 2])
    def test_decode_bleu(self, nbest: int, effective_order: bool):
        metric_type = "bleu"
        metric_cls = get_metric(metric_type)
        decoder = DecoderAggregateMBR(
            DecoderAggregateMBR.Config(),
            metric_cls(metric_cls.Config(effective_order=effective_order)),
        )
        for i, (hyps, refs) in enumerate(zip(HYPOTHESES, REFERENCES)):
            output = decoder.decode(hyps, refs, nbest=nbest, reference_lprobs=None)
            assert output.idx[0] == BEST_INDICES_BLEU[i]
            assert output.sentence[0] == BEST_SENTENCES_BLEU[i]
            assert len(output.sentence) == min(nbest, len(hyps))
            assert len(output.score) == min(nbest, len(hyps))
            torch.testing.assert_close(
                torch.tensor(output.score[0]),
                SCORES_BLEU[effective_order][i],
                atol=0.0005,
                rtol=1e-4,
            )

            output = decoder.decode(
                hyps,
                refs,
                nbest=1,
                reference_lprobs=torch.Tensor([-2.000]).repeat(len(refs)),
            )
            assert output.idx[0] == BEST_INDICES_BLEU[i]
            assert output.sentence[0] == BEST_SENTENCES_BLEU[i]
            torch.testing.assert_close(
                torch.tensor(output.score[0]),
                SCORES_BLEU[effective_order][i],
                atol=0.0005,
                rtol=1e-4,
            )

    @pytest.mark.parametrize("nbest", [1, 2])
    def test_decode_selector(self, nbest: int, selector: Selector):
        metric_type = "bleu"
        metric_cls = get_metric(metric_type)
        decoder = DecoderAggregateMBR(
            DecoderAggregateMBR.Config(),
            metric_cls(metric_cls.Config(effective_order=True)),
            selector=selector,
        )
        for i, (hyps, refs) in enumerate(zip(HYPOTHESES, REFERENCES)):
            output = decoder.decode(hyps, refs, nbest=nbest, reference_lprobs=None)
            assert len(output.sentence) == min(nbest, len(hyps))
            assert len(output.score) == min(nbest, len(hyps))
            output = decoder.decode(
                hyps,
                refs,
                nbest=nbest,
                reference_lprobs=torch.Tensor([-2.000]).repeat(len(refs)),
            )
            assert len(output.sentence) == min(nbest, len(hyps))
            assert len(output.score) == min(nbest, len(hyps))



import pytest
import torch

from .xcomet import MetricXCOMET

SOURCE = "これはテストです"
HYPOTHESES = [
    "this is a test",
    "another test",
    "this is a fest",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
REFERENCES = [
    "ref",
    "this is a test",
    "producţia de zahăr brut se exprimă în zahăr alb;",
]
SCORES = torch.Tensor(
    [
        [0.97671, 1.00000, 0.49054],
        [0.94399, 0.99120, 0.43007],
        [0.71786, 0.71210, 0.30775],
        [0.21788, 0.22079, 0.61004],
    ]
)


@pytest.mark.skipif(
    not torch.cuda.is_available(), reason="CUDA is not available on this machine."
)
class TestMetricXCOMET:
    def test_score(self, metric_xcomet: MetricXCOMET):
        for i, hyp in enumerate(HYPOTHESES):
            for j, ref in enumerate(REFERENCES):
                assert torch.isclose(
                    SCORES[i, j],
                    torch.tensor(metric_xcomet.score(hyp, ref, SOURCE)),
                    atol=0.0005 / 100,
                )

    def test_scores(self, metric_xcomet: MetricXCOMET):
        hyps = ["another test", "this is a test", "this is an test"]
        refs = ["another test", "this is a fest", "this is a test"]
        srcs = [SOURCE] * len(hyps)

        torch.testing.assert_close(
            metric_xcomet.scores(hyps, refs, srcs).cpu().float(),
            torch.FloatTensor([1.00000, 0.90545, 1.00000]),
            atol=0.0005 / 100,
            rtol=1e-6,
        )
        torch.testing.assert_close(
            metric_xcomet.scores(hyps, sources=srcs).cpu().float(),
            torch.FloatTensor([0.99120, 0.99120, 0.99120]),
            atol=0.0005 / 100,
            rtol=1e-6,
        )
        torch.testing.assert_close(
            metric_xcomet.scores(hyps, references=refs).cpu().float(),
            torch.FloatTensor([1.00000, 0.77420, 1.00000]),
            atol=0.0005 / 100,
            rtol=1e-6,
        )

    def test_expected_scores(self, metric_xcomet: MetricXCOMET):
        expected_scores = metric_xcomet.expected_scores(HYPOTHESES, REFERENCES, SOURCE)
        torch.testing.assert_close(
            expected_scores,
            SCORES.mean(dim=1).to(metric_xcomet.device),
            atol=0.0005 / 100,
            rtol=1e-6,
        )

    def test_corpus_score(self, metric_xcomet: MetricXCOMET):
        hyps = ["another test", "this is a test", "this is an test"]
        refs = ["another test", "this is a fest", "this is a test"]
        srcs = [SOURCE] * len(hyps)
        assert torch.isclose(
            torch.tensor(metric_xcomet.corpus_score(hyps, refs, srcs)),
            torch.tensor(0.96848),
        )
        assert torch.isclose(
            torch.tensor(metric_xcomet.corpus_score(hyps, sources=srcs)),
            torch.tensor(0.99120),
        )
        assert torch.isclose(
            torch.tensor(metric_xcomet.corpus_score(hyps, references=refs)),
            torch.tensor(0.92473),
        )



import torch

from .cometkiwi import MetricCOMETkiwi

HYPOTHESES = [
    "this is a test",
    "another test",
    "this is a fest",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
SOURCES = ["これはテストです"] * len(HYPOTHESES)

SCORES = torch.Tensor([0.86415, 0.83704, 0.65335, 0.29771])


class TestMetricCOMETQE:
    def test_score(self, metric_cometkiwi: MetricCOMETkiwi):
        for i, (hyp, src) in enumerate(zip(HYPOTHESES, SOURCES)):
            assert torch.isclose(
                SCORES[i],
                torch.tensor(metric_cometkiwi.score(hyp, src)),
                atol=0.0005 / 100,
            )

    def test_scores(self, metric_cometkiwi: MetricCOMETkiwi):
        scores = metric_cometkiwi.scores(HYPOTHESES, SOURCES)
        torch.testing.assert_close(
            scores,
            SCORES.to(metric_cometkiwi.device),
            atol=0.0005 / 100,
            rtol=1e-6,
        )

    def test_corpus_score(self, metric_cometkiwi: MetricCOMETkiwi):
        hyps = [
            "this is a test",
            "another test",
            "this is a fest",
            "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
        ]
        assert torch.isclose(
            torch.tensor(metric_cometkiwi.corpus_score(hyps, SOURCES)),
            torch.tensor(0.66306),
        )



import multiprocessing
import sys

import pytest
import torch

from .bleu import MetricBLEU

HYPOTHESES = [
    "this is a test",
    "another test",
    "this is a fest",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
REFERENCES = [
    "ref",
    "this is a test",
    "producţia de zahăr brut se exprimă în zahăr alb;",
]
SCORES_EFFECTIVE_ORDER = {
    True: torch.Tensor(
        [
            [0.0, 100.0, 0.0],
            [0.0, 18.394, 0.0],
            [0.0, 59.460, 0.0],
            [0.0, 0.0, 8.493],
        ]
    ),
    False: torch.Tensor(
        [
            [0.0, 100.0, 0.0],
            [0.0, 0.0, 0.0],
            [0.0, 59.460, 0.0],
            [0.0, 0.0, 8.493],
        ]
    ),
}
EXPECTED_SCORES_AGGREGATED = {
    True: torch.Tensor([25.96003, 6.44121, 20.31482, 5.13217]),
    False: torch.Tensor([25.96003, 0.0, 20.31482, 5.13217]),
}


class TestMetricBLEU:
    @pytest.mark.parametrize("effective_order", [True, False])
    def test_score(self, effective_order: bool):
        metric = MetricBLEU(MetricBLEU.Config(effective_order=effective_order))
        for i, hyp in enumerate(HYPOTHESES):
            for j, ref in enumerate(REFERENCES):
                assert torch.isclose(
                    SCORES_EFFECTIVE_ORDER[effective_order][i, j],
                    torch.tensor(metric.score(hyp, ref)),
                    atol=0.0005,
                )

    @pytest.mark.parametrize("effective_order", [True, False])
    def test_expected_scores(self, effective_order: bool):
        metric = MetricBLEU(MetricBLEU.Config(effective_order=effective_order))
        expected_scores = metric.expected_scores(HYPOTHESES, REFERENCES)
        torch.testing.assert_close(
            expected_scores,
            SCORES_EFFECTIVE_ORDER[effective_order].mean(dim=1),
            atol=0.0005,
            rtol=1e-4,
        )

    def test_expected_scores_ja(self):
        metric = MetricBLEU(
            MetricBLEU.Config(tokenize="ja-mecab", effective_order=True)
        )
        hyps = ["ありがとうございます", "どうも"]
        refs = ["ありがとう", "どうもありがとうございます"]
        expected_scores = metric.expected_scores(hyps, refs)
        torch.testing.assert_close(
            expected_scores, torch.Tensor([49.5846, 2.4894]), atol=0.0005, rtol=1e-4
        )
        if sys.platform == "linux":
            default_method = multiprocessing.get_start_method()
            multiprocessing.set_start_method("spawn", force=True)
            expected_scores = metric.expected_scores(hyps, refs)
            torch.testing.assert_close(
                expected_scores, torch.Tensor([49.5846, 2.4894]), atol=0.0005, rtol=1e-4
            )
            multiprocessing.set_start_method(default_method, force=True)

    def test_scores(self):
        hyps = [
            "this is a test",
            "another test",
            "this is a fest",
            "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
        ]
        refs = [
            "this is a test",
            "ref",
            "this is a test",
            "producţia de zahăr brut se exprimă în zahăr alb;",
        ]

        metric = MetricBLEU(MetricBLEU.Config())
        torch.testing.assert_close(
            metric.scores(hyps, refs),
            torch.tensor([metric.score(h, r) for h, r in zip(hyps, refs)]),
        )

    def test_corpus_score(self):
        hyps = [
            "this is a test",
            "another test",
            "this is a fest",
            "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
        ]
        refs = [
            "this is a test",
            "ref",
            "this is a test",
            "producţia de zahăr brut se exprimă în zahăr alb;",
        ]

        metric = MetricBLEU(MetricBLEU.Config())
        assert torch.isclose(
            torch.tensor(metric.corpus_score(hyps, refs)), torch.tensor(22.41424)
        )

    @pytest.mark.parametrize("effective_order", [True, False])
    def test_expected_scores_reference_aggregation(self, effective_order: bool):
        metric = MetricBLEU(MetricBLEU.Config(effective_order=effective_order))
        expected_scores = metric.expected_scores_reference_aggregation(
            HYPOTHESES, REFERENCES
        )
        torch.testing.assert_close(
            expected_scores,
            EXPECTED_SCORES_AGGREGATED[effective_order],
            atol=0.0005,
            rtol=1e-4,
        )

        expected_scores = metric.expected_scores_reference_aggregation(
            HYPOTHESES,
            REFERENCES,
            reference_lprobs=torch.Tensor([-2.000]).repeat(len(REFERENCES)),
        )
        torch.testing.assert_close(
            expected_scores,
            EXPECTED_SCORES_AGGREGATED[effective_order],
            atol=0.0005,
            rtol=1e-4,
        )

    def test_expected_scores_reference_aggregation_empty_inputs(self):
        metric = MetricBLEU(MetricBLEU.Config(effective_order=True))
        hyps = ["thank you", ""]
        refs = ["thank you so much", "", "thank you.", "thank you", ""]
        expected_scores = metric.expected_scores_reference_aggregation(hyps, refs)
        torch.testing.assert_close(
            expected_scores, torch.Tensor([60.0, 0.0]), atol=0.0005, rtol=1e-4
        )

        expected_scores = metric.expected_scores_reference_aggregation(
            hyps, refs, reference_lprobs=torch.Tensor([-2.000]).repeat(len(refs))
        )
        torch.testing.assert_close(
            expected_scores, torch.Tensor([60.0, 0.0]), atol=0.0005, rtol=1e-4
        )



import pytest
import torch

from . import utils


@pytest.mark.skipif(
    not torch.cuda.is_available(), reason="CUDA is not available on this machine."
)
def test_to_device():
    device = torch.device("cuda:0")

    for x in [1, 1.0, "a", True]:
        assert utils.to_device(x, device) == x

    assert utils.to_device(torch.ones(1), device).device == device
    assert utils.to_device({"a": torch.ones(1)}, device)["a"].device == device
    assert utils.to_device([torch.ones(1)], device)[0].device == device
    assert utils.to_device((torch.ones(1),), device)[0].device == device



#!/usr/bin/env python3

import enum
import json
import logging
import os
import sys
from argparse import Namespace
from dataclasses import asdict, dataclass
from typing import Sequence

import simple_parsing
from simple_parsing import choice, field, flag
from simple_parsing.wrappers import dataclass_wrapper

from mbrs import registry
from mbrs.args import ArgumentParser
from mbrs.metrics import Metric, get_metric
from mbrs.metrics.base import MetricReferenceless

simple_parsing.parsing.logger.setLevel(logging.ERROR)
dataclass_wrapper.logger.setLevel(logging.ERROR)

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(name)s | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=os.environ.get("LOGLEVEL", "INFO").upper(),
    stream=sys.stdout,
)
logger = logging.getLogger(__name__)


class Format(enum.Enum):
    plain = "plain"
    json = "json"


@dataclass
class CommonArguments:
    """Common arguments."""

    # Hypotheses file.
    hypotheses: str = field(positional=True)
    # Sources file.
    sources: str | None = field(default=None, alias=["-s"])
    # References file.
    references: str | None = field(default=None, alias=["-r"])
    # Output format.
    format: Format = choice(Format, default=Format.json)
    # Type of the metric.
    metric: str = choice(*registry.get_registry("metric").keys(), default="bleu")
    # No verbose information and report.
    quiet: bool = flag(default=False)
    # Number of digits for values of float point.
    width: int = field(default=1, alias=["-w"])


def get_argparser(args: Sequence[str] | None = None) -> ArgumentParser:
    meta_parser = ArgumentParser(add_help=False, add_config_path_arg=True)
    meta_parser.add_arguments(CommonArguments, "common")
    for _field in meta_parser._wrappers[0].fields:
        _field.required = False
    known_args, _ = meta_parser.parse_known_args(args=args)
    metric_type = get_metric(known_args.common.metric)

    parser = ArgumentParser(add_help=True, add_config_path_arg=True)
    parser.add_arguments(CommonArguments, "common")
    parser.add_arguments(metric_type.Config, "metric", prefix="metric.")
    return parser


def format_argparser() -> ArgumentParser:
    parser = get_argparser()
    parser.preprocess_parser()
    return parser


def main(args: Namespace) -> None:
    with open(args.common.hypotheses, mode="r") as f:
        hypotheses = f.readlines()
    num_sents = len(hypotheses)

    sources = None
    if args.common.sources is not None:
        with open(args.common.sources, mode="r") as f:
            sources = f.readlines()
        assert num_sents == len(sources)

    references = None
    if args.common.references is not None:
        with open(args.common.references, mode="r") as f:
            references = f.readlines()
        assert num_sents == len(references)

    metric_type = get_metric(args.common.metric)
    metric: Metric = metric_type(args.metric)

    if isinstance(metric, MetricReferenceless):
        assert sources is not None
        corpus_score = metric.corpus_score(hypotheses, sources=sources)
    else:
        assert references is not None
        corpus_score = metric.corpus_score(hypotheses, references, sources)

    score = {
        "name": args.common.metric,
        "score": float(f"{corpus_score:.{args.common.width}f}"),
        "metric_cfg": asdict(args.metric),
    }

    if args.common.format == Format.plain:
        print(score["score"])
    elif args.common.format == Format.json:
        print(json.dumps(score, ensure_ascii=False, indent=2))


def cli_main():
    args = get_argparser().parse_args()
    main(args)


if __name__ == "__main__":
    cli_main()



from __future__ import annotations

import concurrent.futures
import itertools
import math
from dataclasses import dataclass

from sacrebleu.metrics.ter import TER
from torch import Tensor

from mbrs import timer

from . import Metric, register


@register("ter")
class MetricTER(Metric):
    """TER metric class."""

    HIGHER_IS_BETTER: bool = False

    @dataclass
    class Config(Metric.Config):
        """TER metric configuration.

        - normalized (bool): Enable character normalization.
          By default, normalizes a couple of things such as newlines being stripped,
          retrieving XML encoded characters, and fixing tokenization for punctuation.
          When 'asian_support' is enabled, also normalizes specific Asian (CJK)
          character sequences, i.e. split them down to the character level.
        - no_punct (bool): Remove punctuation. Can be used in conjunction with
          'asian_support' to also remove typical punctuation markers in Asian languages
          (CJK).
        - asian_support (bool): Enable special treatment of Asian characters.
          This option only has an effect when 'normalized' and/or 'no_punct' is enabled.
          If 'normalized' is also enabled, then Asian (CJK) characters are split down to
          the character level. If 'no_punct' is enabled alongside 'asian_support',
          specific unicode ranges for CJK and full-width punctuations are also removed.
        - case_sensitive (bool): If `True`, does not lowercase sentences.
        - num_workers (int): Number of workers for multiprocessing.
        """

        normalized: bool = False
        no_punct: bool = False
        asian_support: bool = False
        case_sensitive: bool = False
        num_workers: int = 8

    cfg: Config

    def __init__(self, cfg: MetricTER.Config):
        super().__init__(cfg)
        self.scorer = TER(
            normalized=cfg.normalized,
            no_punct=cfg.no_punct,
            asian_support=cfg.asian_support,
            case_sensitive=cfg.case_sensitive,
        )

    def score(self, hypothesis: str, reference: str, *_, **__) -> float:
        """Calculate the score of the given hypothesis.

        Args:
            hypothesis (str): Hypothesis.
            reference (str): Reference.

        Returns:
            float: The score of the given hypothesis.
        """
        return self.scorer.sentence_score(hypothesis, [reference]).score

    def scores(self, hypotheses: list[str], references: list[str], *_, **__) -> Tensor:
        """Calculate the scores of the given hypotheses.

        Args:
            hypotheses (list[str]): N hypotheses.
            references (list[str]): N references.

        Returns:
            Tensor: The N scores of the given hypotheses.
        """
        with concurrent.futures.ProcessPoolExecutor(
            max_workers=self.cfg.num_workers
        ) as executor:
            with timer.measure("score") as t:
                t.set_delta_ncalls(len(hypotheses))
                return Tensor(
                    list(
                        executor.map(
                            self.score,
                            hypotheses,
                            references,
                            chunksize=math.ceil(len(hypotheses) / self.cfg.num_workers),
                        )
                    )
                )

    def pairwise_scores(
        self, hypotheses: list[str], references: list[str], *_, **__
    ) -> Tensor:
        """Calculate the pairwise scores.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.

        Returns:
            Tensor: Score matrix of shape `(H, R)`, where `H` is the number
              of hypotheses and `R` is the number of references.
        """
        with concurrent.futures.ProcessPoolExecutor(
            max_workers=self.cfg.num_workers
        ) as executor:
            with timer.measure("score") as t:
                t.set_delta_ncalls(len(hypotheses) * len(references))

                return Tensor(
                    list(
                        executor.map(
                            self.score,
                            *zip(*itertools.product(hypotheses, references)),
                            chunksize=len(hypotheses),
                        )
                    )
                ).view(len(hypotheses), len(references))

    def corpus_score(
        self, hypotheses: list[str], references: list[str], *_, **__
    ) -> float:
        """Calculate the corpus-level score.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.

        Returns:
            float: The corpus score.
        """
        return self.scorer.corpus_score(hypotheses, [references]).score



import pytest

from . import registry


def test_registry():
    register, get_cls = registry.setup("test")
    REGISTRY = registry.REGISTRIES["test"]
    assert "mock1" not in REGISTRY

    @register("mock1")
    class MockClass1: ...

    assert "mock1" in REGISTRY
    assert issubclass(REGISTRY["mock1"], MockClass1)
    assert issubclass(get_cls("mock1"), MockClass1)

    @register("mock2")
    class MockClass2: ...

    assert "mock2" in REGISTRY
    assert issubclass(REGISTRY["mock2"], MockClass2)
    assert issubclass(get_cls("mock2"), MockClass2)

    assert registry.get_registry("test") == {"mock1": MockClass1, "mock2": MockClass2}

    with pytest.raises(ValueError):

        @register("mock1")
        class MockClass3: ...



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import torch
from comet import download_model, load_from_checkpoint
from comet.models import XCOMETMetric
from torch import Tensor

from mbrs import timer, utils

from . import Metric, register


@register("xcomet")
class MetricXCOMET(Metric):
    """XCOMET metric class."""

    scorer: XCOMETMetric

    @dataclass
    class Config(Metric.Config):
        """XCOMET metric configuration.

        - model (str): Model name or path.
        - batch_size (int): Batch size.
        - fp16 (bool): Use float16 for the forward computation.
        - bf16 (bool): Use bfloat16 for the forward computation.
        - cpu (bool): Use CPU for the forward computation.
        """

        model: str = "Unbabel/XCOMET-XL"
        batch_size: int = 8
        fp16: bool = False
        bf16: bool = False
        cpu: bool = False

    def __init__(self, cfg: MetricXCOMET.Config):
        self.cfg = cfg
        self.scorer = load_from_checkpoint(download_model(cfg.model))
        self.scorer.eval()
        for param in self.scorer.parameters():
            param.requires_grad = False

        if not cfg.cpu and torch.cuda.is_available():
            if cfg.fp16:
                self.scorer = self.scorer.half()
            elif cfg.bf16:
                self.scorer = self.scorer.bfloat16()
            self.scorer = self.scorer.cuda()

    @property
    def device(self) -> torch.device:
        """Returns the device of the model."""
        return self.scorer.device

    def score(
        self,
        hypothesis: str,
        reference: Optional[str] = None,
        source: Optional[str] = None,
    ) -> float:
        """Calculate the score of the given hypothesis.

        Args:
            hypothesis (str): A hypothesis.
            reference (str, optional): A reference.
            source (str, optional): A source.

        Returns:
            float: The score of the given hypothesis.
        """
        inputs = {"mt": hypothesis}
        if reference is not None:
            inputs["ref"] = reference
        if source is not None:
            inputs["src"] = source

        batch = self.scorer.prepare_for_inference([inputs])
        batch = utils.to_device(batch, self.device)
        model_output = self.scorer.predict_step(batch)
        return model_output.scores.item()

    def scores(
        self,
        hypotheses: list[str],
        references: Optional[list[str]] = None,
        sources: Optional[list[str]] = None,
    ) -> Tensor:
        """Calculate the scores of the given hypothesis.

        Args:
            hypotheses (list[str]): N hypotheses.
            references (list[str], optional): N references.
            sources (list[str], optional): N sources.

        Returns:
            Tensor: The N scores of the given hypotheses.
        """
        inputs = [{"mt": hyp} for hyp in hypotheses]
        if references is not None:
            for d, ref in zip(inputs, references):
                d["ref"] = ref
        if sources is not None:
            for d, src in zip(inputs, sources):
                d["src"] = src

        scores = []
        with timer.measure("score") as t:
            t.set_delta_ncalls(len(inputs))
            for i in range(0, len(inputs), self.cfg.batch_size):
                batch = self.scorer.prepare_for_inference(
                    inputs[i : i + self.cfg.batch_size]
                )
                batch = utils.to_device(batch, self.device)
                model_output = self.scorer.predict_step(batch)
                scores.append(model_output.scores)
        return torch.cat(scores).view(len(hypotheses))

    def pairwise_scores(
        self, hypotheses: list[str], references: list[str], source: Optional[str] = None
    ) -> Tensor:
        """Calculate the pairwise scores.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.

        Returns:
            Tensor: Score matrix of shape `(H, R)`, where `H` is the number
              of hypotheses and `R` is the number of references.
        """
        data = [
            {"src": source, "mt": hyp, "ref": ref}
            for hyp in hypotheses
            for ref in references
        ]
        scores = []
        with timer.measure("score") as t:
            t.set_delta_ncalls(len(data))
            for i in range(0, len(data), self.cfg.batch_size):
                batch = self.scorer.prepare_for_inference(
                    data[i : i + self.cfg.batch_size]
                )
                batch = utils.to_device(batch, self.device)
                model_output = self.scorer.predict_step(batch)
                scores.append(model_output.scores)
        return torch.cat(scores).view(len(hypotheses), len(references))

    def corpus_score(
        self,
        hypotheses: list[str],
        references: Optional[list[str]] = None,
        sources: Optional[list[str]] = None,
    ) -> float:
        """Calculate the corpus-level score.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str], optional): References.
            sources (list[str], optional): Sources.

        Returns:
            float: The corpus score.
        """
        scores = []
        for i in range(0, len(hypotheses), self.cfg.batch_size):
            scores.append(
                self.scores(
                    hypotheses[i : i + self.cfg.batch_size],
                    references[i : i + self.cfg.batch_size] if references is not None else None,
                    sources[i : i + self.cfg.batch_size] if sources is not None else None,
                )
                .float()
                .cpu()
            )
        return torch.cat(scores).mean().item()
